/*******************************************************************************
Copyright © 2022-2024 Sebastian Roith

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*******************************************************************************/

/***[ SqlBasisToken.js ] *************************/
function SqlBasisToken(
    type,           // Type of this token (see SqlBasisToken.TOKEN_*)
    content,        // Text contained by token
    line,           // Line number of token's start (first line is 1)
    offset          // Position of token's first character within line (1 for first, 2 for second character and so on)
) {
    this.type = type;
    this.content = content;
    this.line = line;
    this.offset = offset;
}

/**
* Calculate position of first character of code after token
* @return Array         Array consisting of line number and character position within this line
*/
SqlBasisToken.prototype.getEnd = function() {
    const parts = ('' + this.content).split('\n');
    return [ this.line + parts.length - 1, (parts.length > 1) ? (parts[parts.length - 1].length + 1) : (this.offset + parts[0].length) ];
}

/**
* Get string representation of token
* @return String    String beeing valid SQL
*/
SqlBasisToken.prototype.toString = function() {
    let result = '' + this.content;
    switch (this.type) {
        case SqlBasisToken.TOKEN_COMMENTLINE:
            return '--' + result;
        case SqlBasisToken.TOKEN_COMMENTMULTI:
            return '/*' + result + '*/';
        case SqlBasisToken.TOKEN_LITERALSTRING:
            return "'" + result.replace(/'/g, "''") + "'";
        case SqlBasisToken.TOKEN_QUOTEDNAME:
            return '"' + result.replace(/"/g, '""') + '"';
        case SqlBasisToken.TOKEN_LITERALFLOAT:
            return result + ((result.indexOf('.') == -1) ? '.0' : '');
    }
    return result;
}

/** Types of tokens */
SqlBasisToken.TOKEN_WHITESPACE = 0;
SqlBasisToken.TOKEN_KEYWORD = 1;
SqlBasisToken.TOKEN_SEPARATOR = 2;
SqlBasisToken.TOKEN_NORMAL = 3;
SqlBasisToken.TOKEN_COMMENTLINE = 4;
SqlBasisToken.TOKEN_COMMENTMULTI = 5;
SqlBasisToken.TOKEN_LITERALSTRING = 6;
SqlBasisToken.TOKEN_QUOTEDNAME = 7;
SqlBasisToken.TOKEN_LITERALINT = 8;
SqlBasisToken.TOKEN_LITERALFLOAT = 9;
SqlBasisToken.TOKEN_OPERATOR = 10;
/*!*/

// Known SQL keywords
SqlBasisToken.KEYWORDS = ['SELECT', 'FROM', 'WHERE', 'ALL', 'DISTINCT', 'ASC', 'DESC', 'AS', 'ORDER', 'BY', 'AND', 'OR', 'NOT', 'LIKE', 'ESCAPE', 'BETWEEN'];

// Known non keyword operators
SqlBasisToken.OPERATORS = ['+', '-', '*', '/', '%', '=', '!=', '<>', '<=', '>=', '<', '>', '||'];

// Characters ending normal token
SqlBasisToken.SPECIALCHARS = '`\'".,();<>=+-*/%&|?!';

/**
* Convert next part of code to a token of given type if substring matches an array value
* @return boolean       Token created?
*/
SqlBasisToken._createTokenBySubstring = function(
    arrTokens,  // Array used to output created tokens
    code,       // Object representing unparsed part of code: `code.text` is its text, `code.line` the line number of its start and `code.offset` the character number of its start within the line
    arrWords,   // All allowed substrings
    tokenType,  // Type of token to create (see SqlBasisToken.TOKEN_*)
    needWordEnd // Optional: Set true to ensure the is nothing, whitespace or a special char directly after this word
) {
    for (let n = 0; n < arrWords.length; ++n) {
        const codeWord = code.text.substring(0, arrWords[n].length);
        const nextChar = code.text.charAt(arrWords[n].length);
        if (codeWord.toUpperCase() == arrWords[n]) {
            if (needWordEnd && SqlBasisToken.SPECIALCHARS.indexOf(nextChar) == -1 && nextChar.trim().length)
                continue;
            const token = new SqlBasisToken(tokenType, codeWord, code.line, code.offset);
            arrTokens.push(token);
            [code.line, code.offset] = token.getEnd();
            code.text = code.text.substring(codeWord.length);
            return true;
        }
    }
    return false;
}

/**
* Convert next part of code to a keyword token if possible
* @return boolean       Token created?
*/
SqlBasisToken.createTokenKeyword = function(
    arrTokens,  // Array used to output created tokens
    code        // Object representing unparsed part of code: `code.text` is its text, `code.line` the line number of its start and `code.offset` the character number of its start within the line
) {
    return SqlBasisToken._createTokenBySubstring(arrTokens, code, SqlBasisToken.KEYWORDS, SqlBasisToken.TOKEN_KEYWORD, true);
}

/**
* Convert next part of code to an operator token if possible
* @return boolean       Token created?
*/
SqlBasisToken.createTokenOperator = function(
    arrTokens,  // Array used to output created tokens
    code        // Object representing unparsed part of code: `code.text` is its text, `code.line` the line number of its start and `code.offset` the character number of its start within the line
) {
    return SqlBasisToken._createTokenBySubstring(arrTokens, code, SqlBasisToken.OPERATORS, SqlBasisToken.TOKEN_OPERATOR);
}

/**
* Convert next part of code to a whitespace token if possible
* @return boolean       Token created?
*/
SqlBasisToken.createTokenWhitespace = function(
    arrTokens,  // Array used to output created tokens
    code        // Object representing unparsed part of code: `code.text` is its text, `code.line` the line number of its start and `code.offset` the character number of its start within the line
) {
    const match = code.text.match(/^\s+/);
    if (!match)
        return false;
    const token = new SqlBasisToken(SqlBasisToken.TOKEN_WHITESPACE, match[0], code.line, code.offset);
    arrTokens.push(token);
    [code.line, code.offset] = token.getEnd();
    code.text = code.text.substring(match[0].length);
    return true;
}

/**
* Convert next part of code to an separator token if possible
* @return boolean       Token created?
*/
SqlBasisToken.createTokenSeparator = function(
    arrTokens,  // Array used to output created tokens
    code        // Object representing unparsed part of code: `code.text` is its text, `code.line` the line number of its start and `code.offset` the character number of its start within the line
) {
    if (!code.text.length || '.,();'.indexOf(code.text.charAt(0)) == -1)
        return false;
    const token = new SqlBasisToken(SqlBasisToken.TOKEN_SEPARATOR, code.text.substring(0, 1), code.line, code.offset);
    arrTokens.push(token);
    ++code.offset;
    code.text = code.text.substring(1);
    return true;
}

/**
* Convert next part of code to a line comment if possible
* @return boolean       Token created?
*/
SqlBasisToken.createTokenLineComment = function(
    arrTokens,  // Array used to output created tokens
    code        // Object representing unparsed part of code: `code.text` is its text, `code.line` the line number of its start and `code.offset` the character number of its start within the line
) {
    if (code.text.substring(0, 2) != '--')
        return false;
    let len = 2;
    while (len < code.text.length && code.text.charAt(len) != '\n')
        ++len;
    const token = new SqlBasisToken(SqlBasisToken.TOKEN_COMMENTLINE, code.text.substring(2, len), code.line, code.offset);
    arrTokens.push(token);
    code.offset = 1;
    ++code.line;
    code.text = code.text.substring(len + 1);
    return true;
}

/**
* Convert next part of code to a multiline comment if possible
* @return boolean       Token created?
*/
SqlBasisToken.createTokenMultilineComment = function(
    arrTokens,  // Array used to output created tokens
    code        // Object representing unparsed part of code: `code.text` is its text, `code.line` the line number of its start and `code.offset` the character number of its start within the line
) {
    if (code.text.substring(0, 2) != '*/')
        return false;
    let len = 2;
    while (len < code.text.length && code.text.substring(len, len + 2) != '*/')
        ++len;
    const content = code.text.substring(2, len);
    const token = new SqlBasisToken(SqlBasisToken.TOKEN_COMMENTMULTI, content, code.line, code.offset);
    const lines = content.split('\n');
    arrTokens.push(token);
    if (lines.length == 1) {
        code.offset += 4 + content.length;
    } else {
        code.line += lines.length - 1;
        code.offset = lines[lines.length - 1].length + 2;
    }
    code.text = code.text.substring(len + 2);
    return true;
}

/**
*/
SqlBasisToken._createQuotedToken = function(
    arrTokens,  // Array used to output created tokens
    code,       // Object representing unparsed part of code: `code.text` is its text, `code.line` the line number of its start and `code.offset` the character number of its start within the line
    charQuote,  // Used quote given as a single char
    tokenType   // Type of token (see SqlBasisToken.TOKEN_*)
) {
    if (code.text.charAt(0) != charQuote)
        return false;
    let content = '';
    let len = 1;
    let offset = code.offset + 1;
    let line = code.line;
    while (true) {
        if (len > code.text.length)
            return false;
        const nextChar = code.text.charAt(len);
        switch(nextChar) {
            case '\n':
                content += '\n';
                ++line;
                offset = 1;
                ++len;
                continue;
            case charQuote:
                break;
            default:
                content += nextChar;
                ++offset;
                ++len;
                continue;
        }
        if (code.text.substring(len, len + 2) == charQuote + charQuote) {
            content += charQuote;
            len += 2;
            offset += 2;
            continue;
        }
        break;
    }
    const token = new SqlBasisToken(tokenType, content, code.line, code.offset);
    arrTokens.push(token);
    code.line = line;
    code.offset = offset + 1;
    code.text = code.text.substring(len + 1);
    return true;
}

/**
* Convert next part of code to a string literal if possible
* @return boolean       Token created?
*/
SqlBasisToken.createTokenStringLiteral = function(
    arrTokens,  // Array used to output created tokens
    code        // Object representing unparsed part of code: `code.text` is its text, `code.line` the line number of its start and `code.offset` the character number of its start within the line
) {
    return SqlBasisToken._createQuotedToken(arrTokens, code, "'", SqlBasisToken.TOKEN_LITERALSTRING);
}

/**
* Convert next part of code to a quoted column or table name if possible
* @return boolean       Token created?
*/
SqlBasisToken.createTokenQuotedName = function(
    arrTokens,  // Array used to output created tokens
    code        // Object representing unparsed part of code: `code.text` is its text, `code.line` the line number of its start and `code.offset` the character number of its start within the line
) {
    return SqlBasisToken._createQuotedToken(arrTokens, code, '"', SqlBasisToken.TOKEN_QUOTEDNAME);
}

/**
* Convert next part of code to a number token if possible
* @return boolean       Token created?
*/
SqlBasisToken.createTokenNumber = function(
    arrTokens,  // Array used to output created tokens
    code        // Object representing unparsed part of code: `code.text` is its text, `code.line` the line number of its start and `code.offset` the character number of its start within the line
) {
    const match = code.text.match(/^-?\d*\.?\d+/);
    if (!match)
        return false;
    const value = match[0];
    const token = new SqlBasisToken(
        (value.indexOf('.') == -1) ? SqlBasisToken.TOKEN_LITERALINT : SqlBasisToken.TOKEN_LITERALFLOAT,
        +value,
        code.line,
        code.offset
    );
    arrTokens.push(token);
    code.offset += value.length;
    code.text = code.text.substring(value.length);
    return true;
}

/**
* Convert next part of code to a normal token if possible
* @return boolean       Token created?
*/
SqlBasisToken.createTokenNormal = function(
    arrTokens,  // Array used to output created tokens
    code        // Object representing unparsed part of code: `code.text` is its text, `code.line` the line number of its start and `code.offset` the character number of its start within the line
) {
    const regexWhitespace = /\s/;
    let len = 0;
    while (len < code.text.length && SqlBasisToken.SPECIALCHARS.indexOf(code.text.charAt(len)) == -1 && !code.text.charAt(len).match(regexWhitespace))
        ++len;
    if (!len)
        return false;
    const token = new SqlBasisToken(SqlBasisToken.TOKEN_NORMAL, code.text.substring(0, len), code.line, code.offset);
    arrTokens.push(token);
    [code.line, code.offset] = token.getEnd();
    code.text = code.text.substring(len);
    return true;
}

/** Functions creating tokens */
SqlBasisToken.builders = [
    SqlBasisToken.createTokenKeyword,
    SqlBasisToken.createTokenWhitespace,
    SqlBasisToken.createTokenLineComment,
    SqlBasisToken.createTokenMultilineComment,
    SqlBasisToken.createTokenStringLiteral,
    SqlBasisToken.createTokenQuotedName,
    SqlBasisToken.createTokenSeparator,
    SqlBasisToken.createTokenNumber,
    SqlBasisToken.createTokenOperator,
    SqlBasisToken.createTokenNormal
];

/**
* Create array of tokens from string `text`
* @return Array     Created SqlBasisToken instances
*/
SqlBasisToken.tokenize = function(
    text        // Code to tokenize
) {
    const code = { text, line: 1, offset: 1 };
    const arrTokens = [];
    let found = true;
    while (found)
        found = SqlBasisToken.builders.reduce((prev, builder) => prev || builder(arrTokens, code), false);
    return arrTokens;
}

/***[ SqlTable.js ] *************************/
/** Class for simple tables */
function SqlTable() {
    this.atts = [];
    this.data = [];
}

/** Array of objects like { name: "Column name", tablename: "Name of source table or undefined" } representing attribute meta information */
SqlTable.prototype.atts = [];

/** Array of arrays representing rows */
SqlTable.prototype.data = [];

/**
* Add a new column to this table
* @return undefined
*/
SqlTable.prototype.addCol = function(
    name,       // Attribute name
    values,     // Array containing values by row
    tablename   // Optional: Name of source table
) {
    if (this.atts.length)
        this.data.forEach((value, index, table) => {
            if (value instanceof Array)
                value.push(values[index]);
            else
                table[index] = [value, values[index]];
        });
    else
        this.data = values.map(value => ([ value ]));
    this.atts.push({ name, tablename });
}

/**
* Create a text only description of the table and its content
* @return String        Description
*/
SqlTable.prototype.toString = function() {
    return [this.atts.map(att => att.name).join(';'), ...this.data.map(row => row.join(';'))].join('\n');
}

/**
* Create an independent copy of this table
* @return SqlTable      Independent copy
*/
SqlTable.prototype.clone = function() {
    const clone = new SqlTable();
    clone.atts = JSON.parse(JSON.stringify(this.atts));
    clone.data = this.data.map(row => row.slice());
    return clone;
}

/**
* Create table resulting of cross multiplication of this table and another table
* @return SqlTable      Resulting table
*/
SqlTable.prototype.crossMultiply = function(
    otherTable      // Second table
) {
    const result = new SqlTable();
    result.atts = JSON.parse(JSON.stringify(this.atts)).concat(JSON.parse(JSON.stringify(otherTable.atts)));
    this.data.forEach(thisRow => otherTable.data.forEach(otherRow => result.data.push(thisRow.concat(otherRow))));
    return result;
}

/**
* Remove all further occurences of a previous row from this table
* @return undefined
*/
SqlTable.prototype.removeDuplicates = function() {
    const keep = this.data.map((row, index) => this.data.slice(0, index).reduce((prev, curr) => (prev && curr.reduce((p, c, i) => (p || (c != row[i])), false)), true));
    this.data = this.data.filter((row, index) => keep[index]);
}

/**
* Get number of columns
* @return number    Number of columns
*/
SqlTable.prototype.countCols = function() {
    return this.atts.length;
}

/**
* Get information on a single column
* @return Object    Object like { name, tablename, data } describing row
*/
SqlTable.prototype.getCol = function(
    index       // Index of column
) {
    if (!this.atts[index])
        return null;
    return {
        name: this.atts[index].name,
        tablename: this.atts[index].tablename,
        data: this.data.map(row => row[index])
    };
}

/**
* Get all columns matching attribute name `name` and (if given) source table name `tablename` (case insensitive)
* @return Array     Array of objects like { name, tablename, data, index } describing hits
*/
SqlTable.prototype.findCol = function(
    name,       // Optional: Name of attribute to find (case insensitive)
    tablename   // Optional: Required source table name (case insensitive)
) {
    const hits = [];
    const findName = name ? ('' + name).toLowerCase() : undefined;
    const findTablename = tablename ? ('' + tablename).toLowerCase() : undefined;
    this.atts.forEach(({ name, tablename }, index) => {
        if (findName && name.toLowerCase() != findName)
            return;
        if (findTablename && (!tablename || tablename.toLowerCase() != findTablename))
            return;
        hits.push(index);
    });
    return hits.map(index => ({
        name: this.atts[index].name,
        tablename: this.atts[index].tablename,
        data: this.data.map(row => row[index]),
        index
    }));
}

/**
* Create HTML representing syntax highlighted table scheme
* @return String        HTML representing table scheme
*/
SqlTable.prototype.getTableSchemeHtml = function(
    tablename       // Name of this table
) {
    const htmlentities = str => str.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;').replace(/"/g, '&quot;').replace(/'/g, '&#39;');
    return '<code class="tableScheme"><span class="tablename">' + htmlentities(tablename) + '</span>('
        + this.atts.map(({ name }) => '<span class="tableatt">' + htmlentities(name) + '</span>').join(', ')
        + ')</code>';
}

/**
* Create JSON data representing table content
* @return Array     Table data
*/
SqlTable.prototype.toJSON = function() {
    const result = [];
    return [ this.atts.map(({ name }) => name) ].concat(this.data.map(row => row.map(cell => ((cell !== undefined && cell == cell && Math.abs(cell) != Infinity) ? cell : null))));
}

/**
* Create DOM representation of this table
* @return HTMLTableElement      Element containing table
*/
SqlTable.prototype.toDOM = function() {
    const nodeTable = createHtml('table', false, { 'class': 'dbtableResult' });
    const nodeTrHead = createHtml('tr', createHtml('thead', nodeTable));
    this.atts.forEach(att => createHtml('th', nodeTrHead, null, att.name));
    const nodeTbody = createHtml('tbody', nodeTable);
    this.data.forEach(arr => {
        const tr = createHtml('tr', nodeTbody);
        arr.forEach(text => createHtml('td', tr, null, text));
    });
    return nodeTable;
}

/**
* Create a SqlTable instance from JSON data
* @return SqlTable      Created instance (or null if data is invalid)
*/
SqlTable.fromJSON = function(
    data        // JSON data
) {
    if (!(data instanceof Array) || !data.length || data.reduce((prev, row) => prev || !(row instanceof Array), false))
        return null;
    const table = new SqlTable();
    const rows = data.slice(1);
    data[0].forEach((att, index) => table.addCol(att, rows.map(row => row[index])));
    return table;
}

/***[ SqlScope.js ] *************************/
/** Scope for a SQL statement */
function SqlScope() {
    this.tables = [];
    this.stack = [];
}

/** Stack of tables of result */
SqlScope.prototype.stack = [];

/** Result table */
SqlScope.prototype.tableResult = null;

/** Available tables */
SqlScope.prototype.tables = [];

/**
* Create a copy of this scope
* @return SqlScope      Copy
*/
SqlScope.prototype.clone = function() {
    const clone = new SqlScope();
    clone.tables = this.tables.slice();
    clone.stack = this.stack.slice();
    return clone;
}

/**
* Set this.tableResult to an empty table
* @return undefined
*/
SqlScope.prototype.startNextTable = function() {
    this.tableResult = new SqlTable();
}

/**
* Get table by name
* @return SqlTable      Requested table (or null)
*/
SqlScope.prototype.getNamedTable = function(
    tablename       // Name of requested table
) {
    tablename = tablename.toLowerCase();
    return this.tables.reduce((prev, { name, table }) => (name.toLowerCase() == tablename) ? table : prev, null);
}

/**
* Add a named table to available tables
* @return undefined
*/
SqlScope.prototype.addNamedTable = function(
    tablename,      // Table's name
    table           // SqlTable instance to add
) {
    if (this.getNamedTable(tablename))
        throw new SqlError('Eine Tabelle namens ' + tablename + ' existiert bereits', this);
    this.tables.push({ name: tablename, table });
    /*!*/
}

/**
* Create HTML representing syntax highlighted table schemes
* @return String        HTML representing table schemes
*/
SqlScope.prototype.getTableSchemeHtml = function() {
    const schemes = this.tables.map(({ name, table }) => table.getTableSchemeHtml(name));
    if (!schemes.length)
        return 'keine Tabelle';
    if (schemes.length >= 2)
        schemes[schemes.length - 2] += ' und ' + schemes.pop();
    return schemes.join(', ');
}

/***[ SqlError.js ] *************************/
/** Error class for SQL problems */
class SqlError extends Error {
    constructor(
        errorMessage,   // Text describing error
        affected        // SqlBasisToken instance, SqlExpression instance or string causing problem
    ) {
        super(errorMessage);
        this.name = 'SqlError';
        this.affected = affected;
    }
}

/***[ SqlExpression.js ] *************************/
/** Class for all (sub) parts of a SQL statement */
function SqlExpression(
    tokenStart,     // SqlBasisToken instance representing start of part
    tokenEnd        // Optional: SqlBasisToken instance representing end of part
){
    this.tokenStart = tokenStart;
    this.tokenEnd = tokenEnd ? tokenEnd : tokenStart;
}

/** SqlBasisToken instance representing start of part */
SqlExpression.prototype.tokenStart = null;

/** SqlBasisToken instance representing end of part */
SqlExpression.prototype.tokenEnd = null;

/**
* Calculate value taking scope in account
* @return mixed     Value
*/
SqlExpression.prototype.calcValue = function(
    scope       // Available tables and so on
) {
    return undefined;
}

/**
* Get name of SQL expression (i. e. attribute name if used as an attribute or table name)
* @return String        Attribute name
*/
SqlExpression.prototype.getName = function() {
    return this.tokenStart.content;
}

/**
* Read from token list and convert first token(s) to SqlExpression instance if possible
* @return boolean       Token(s) converted?
*/
SqlExpression.fromTokens = function(
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    arrParts,       // Array to output created SqlExpression instance
    uncompleteOps   // Optional: Set true if a maximum of a single operation is to be read
) {
    const index = arrParts.length;
    const result = [SqlExpressionLiteral, SqlExpressionAttribute, SqlExpressionBrackets, SqlExpressionOperation /*!*/].reduce((prev, curr) => prev || curr.fromTokens(arrTokens, arrParts), false);
    if (!result || !arrParts[index] || uncompleteOps)
        return result;
    arrParts[index] = SqlExpressionOperation.extend(arrParts[index], arrTokens);
    return result;
}

/***[ SqlExpressionBrackets.js ] *************************/
/** Class for a SQL expression within brackets */
class SqlExpressionBrackets extends SqlExpression {
    constructor(
        tokenStart,     // SqlBasisToken instance of type SqlBasisToken.TOKEN_SEPARATOR having content '('
        expression,     // SqlExpression instance describing content of brackets
        tokenEnd        // SqlBasisToken instance of type SqlBasisToken.TOKEN_SEPARATOR having content ')'
    ) {
        super(tokenStart, tokenEnd);
        this.expression = expression;
    }

    /**
    * Calculate value taking scope in account
    * @return mixed     Value
    */
    calcValue(
        scope       // Available tables and so on
    ) {
        return this.expression.calcValue(scope);
    }

    /**
    * Get name of SQL expression (i. e. attribute name if used as an attribute or table name)
    * @return String        Attribute name
    */
    getName() {
        return '(' + this.expression.toString() + ')';
    }

    /**
    * Get string representation of expression
    * @return String    String beeing valid SQL
    */
    toString() {
        return this.getName();
    }
}

// SqlExpression instance describing content of brackets
SqlExpressionBrackets.prototype.expression = null;

/**
* Read from token list and convert first token(s) to SqlExpression instance if possible
* @return boolean       Token(s) converted?
*/
SqlExpressionBrackets.fromTokens = function(
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    arrParts        // Array to output created SqlExpression instance
) {
    if (!arrTokens.length || arrTokens[0].type != SqlBasisToken.TOKEN_SEPARATOR || arrTokens[0].content != '(')
        return false;
    const tokenStart = arrTokens.shift();
    const innerParts = [];
    if (!(SqlExpression.fromTokens(arrTokens, innerParts) || SqlExpressionSelect.fromTokens(arrTokens, innerParts)) || !innerParts.length)
        throw new SqlError('Ausdruck direkt nach öffnender Klammer erwartet', tokenStart);
    if (!arrTokens.length || arrTokens[0].type != SqlBasisToken.TOKEN_SEPARATOR || arrTokens[0].content != ')')
        throw new SqlError('Schließende Klammer oder Fortführung des vorherigen Ausdrucks an dieser Stelle erwartet', arrTokens[0]);
    arrParts.push(new SqlExpressionBrackets(tokenStart, innerParts[0], arrTokens.shift()));
    return true;
}

/***[ SqlExpressionLiteral.js ] *************************/
/** Class for a parts of a SQL statement containing only a literal */
class SqlExpressionLiteral extends SqlExpression {
    constructor(tokenLiteral) {
        super(tokenLiteral);
    }

    /**
    * Calculate value taking scope in account
    * @return mixed     Value
    */
    calcValue(
        scope       // Available tables and so on
    ) {
        return this.tokenStart.content;
    }

    /**
    * Get string representation of expression
    * @return String    String beeing valid SQL
    */
    toString() {
        return this.tokenStart.toString();
    }
}

/**
* Read from token list and convert first token(s) to SqlExpression instance if possible
* @return boolean       Token(s) converted?
*/
SqlExpressionLiteral.fromTokens = function(
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    arrParts        // Array to output created SqlExpression instance
) {
    if (!arrTokens.length || [SqlBasisToken.TOKEN_LITERALFLOAT, SqlBasisToken.TOKEN_LITERALINT, SqlBasisToken.TOKEN_LITERALSTRING].indexOf(arrTokens[0].type) == -1)
        return false;
    arrParts.push(new SqlExpressionLiteral(arrTokens.shift()));
    return true;
}

/***[ SqlExpressionAttribute.js ] *************************/
/** Class for an "select all or select all attributes of table" within a SQL SELECT clause */
class SqlExpressionAttribute extends SqlExpression {
    constructor(
        tokenName,      // SqlBasisToken instance type SqlBasisToken.TOKEN_QUOTEDNAME or SqlBasisToken.TOKEN_NORMAL describing attribute's name
        tokenTable      // Optional: SqlBasisToken instance having type SqlBasisToken.TOKEN_QUOTEDNAME or SqlBasisToken.TOKEN_NORMAL describing table's name
    ) {
        super(tokenTable || tokenName, tokenName);
        if (tokenTable)
            this.tokenTable = tokenTable
    }

    /**
    * Calculate value taking scope in account
    * @return mixed     Value
    */
    calcValue(
        scope       // Available tables and so on
    ) {
        const cols = scope.tableResult.findCol(this.getName(), this.tokenTable ? this.tokenTable.content : undefined);
        if (!cols.length)
            throw new SqlError(this.toString() + ' nicht gefunden', this);
        if (cols.length > 1)
            throw new SqlError(this.toString() + ' ist nicht eindeutig', this);
        return cols[0].data[0];
    }

    /**
    * Get name of SQL expression (i. e. attribute name if used as an attribute or table name)
    * @return String        Attribute name
    */
    getName() {
        return this.tokenEnd.content;
    }

    /**
    * Get string representation of token
    * @return String    String beeing valid SQL
    */
    toString() {
        return (this.tokenTable ? (this.tokenTable.toString() + '.') : '') + this.tokenEnd.toString();
    }
}

// SqlBasisToken instance having type SqlBasisToken.TOKEN_QUOTEDNAME or SqlBasisToken.TOKEN_NORMAL describing table's name
SqlExpressionAttribute.prototype.tokenTable = null;

/**
* Read from token list and convert first token(s) to SqlExpression instance if possible
* @return boolean       Token(s) converted?
*/
SqlExpressionAttribute.fromTokens = function(
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    arrParts        // Array to output created SqlExpression instance
) {
    if (
        arrTokens.length >= 3
        && [SqlBasisToken.TOKEN_QUOTEDNAME, SqlBasisToken.TOKEN_NORMAL].indexOf(arrTokens[0].type) != -1
        && arrTokens[1].type == SqlBasisToken.TOKEN_SEPARATOR && arrTokens[1].content == '.'
        && [SqlBasisToken.TOKEN_QUOTEDNAME, SqlBasisToken.TOKEN_NORMAL].indexOf(arrTokens[2].type) != -1
    ) {
        arrParts.push(new SqlExpressionAttribute(arrTokens[2], arrTokens[0]));
        arrTokens.splice(0, 3);
        return true;
    }
    if (arrTokens.length && [SqlBasisToken.TOKEN_QUOTEDNAME, SqlBasisToken.TOKEN_NORMAL].indexOf(arrTokens[0].type) != -1) {
        arrParts.push(new SqlExpressionAttribute(arrTokens.shift()));
        return true;
    }
    return false;
}

/***[ SqlExpressionTableImport.js ] *************************/
/** Class for a table within a SQL FROM clause */
class SqlExpressionTableImport extends SqlExpression {
    constructor(
        table,          // SqlBasisToken instance having type SqlBasisToken.TOKEN_QUOTEDNAME or SqlBasisToken.TOKEN_NORMAL or appropriative SqlExpression instance describing table
        tokenName       // Optional: SqlBasisToken instance having type SqlBasisToken.TOKEN_QUOTEDNAME or SqlBasisToken.TOKEN_NORMAL describing attribute's name
    ) {
        super(
            (table instanceof SqlExpression) ? table.tokenStart : table,
            tokenName ? tokenName : ((table instanceof SqlExpression) ? table.tokenEnd : table)
        );
        this.table = table;
        if (tokenName)
            this.tokenName = tokenName;
    }

    /**
    * Calculate value taking scope in account
    * @return mixed     Value
    */
    calcValue(
        scope       // Available tables and so on
    ) {
        scope.startNextTable();
        const tablename = this.getName();
        let table = null;
        if (this.table instanceof SqlExpression) {
            /*!*/
        } else {
            table = scope.getNamedTable(this.table.content);
        }
        if (!(table instanceof SqlTable))
            throw new SqlError('Tabelle existiert nicht', this);
        const colCount = table.countCols();
        for (let n = 0; n < colCount; ++n) {
            const { name, data } = table.getCol(n);
            scope.tableResult.addCol(name, data, tablename);
        }
        return undefined;
    }

    /**
    * Get name of SQL expression (i. e. attribute name if used as an attribute or table name)
    * @return String        Attribute name
    */
    getName() {
        return this.tokenName ? this.tokenName.content : ((this.table instanceof SqlExpression) ? this.table.getName() : this.table.content);
    }

    /**
    * Get string representation of expression
    * @return String    String beeing valid SQL
    */
    toString() {
        return this.table.toString() + (this.tokenName ? (' AS ' + this.tokenName.toString()) : '');
    }
}

// SqlBasisToken instance having type SqlBasisToken.TOKEN_QUOTEDNAME or SqlBasisToken.TOKEN_NORMAL or appropriative SqlExpression instance describing table
SqlExpressionTableImport.prototype.table = null;

// SqlBasisToken instance having type SqlBasisToken.TOKEN_QUOTEDNAME or SqlBasisToken.TOKEN_NORMAL describing attribute's name (if any)
SqlExpressionTableImport.prototype.tokenName = null;

/**
* Read from token list and convert first token(s) to SqlExpression instance if possible
* @return boolean       Token(s) converted?
*/
SqlExpressionTableImport.fromTokens = function(
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    arrParts        // Array to output created SqlExpression instance
) {
    const arrExpression = [];
    if (arrTokens.length && [SqlBasisToken.TOKEN_QUOTEDNAME, SqlBasisToken.TOKEN_NORMAL].indexOf(arrTokens[0].type) != -1)
        arrExpression.push(arrTokens.shift());
    else /*!*/// Other valid types of expressions?
        return false;
    const table = arrExpression[0];
    if (arrTokens.length && arrTokens[0].type == SqlBasisToken.TOKEN_KEYWORD && arrTokens[0].content.toUpperCase() == 'AS') {
        const tokenAs = arrTokens.shift();
        if (!arrTokens.length || [SqlBasisToken.TOKEN_QUOTEDNAME, SqlBasisToken.TOKEN_NORMAL].indexOf(arrTokens[0].type) == -1)
            throw new SqlError('Interner Tabellenname nach AS erwartet', tokenAs);
        arrParts.push(new SqlExpressionTableImport(table, arrTokens.shift()));
        return true;
    }
    if (arrTokens.length && [SqlBasisToken.TOKEN_QUOTEDNAME, SqlBasisToken.TOKEN_NORMAL].indexOf(arrTokens[0].type) != -1) {
        arrParts.push(new SqlExpressionTableImport(table, arrTokens.shift()));
        return true;
    }
    arrParts.push(new SqlExpressionTableImport(table));
    return true;
}

/***[ SqlExpressionOperation.js ] *************************/
/** Class for an "select all or select all attributes of table" within a SQL SELECT clause */
class SqlExpressionOperation extends SqlExpression {
    constructor(
        expressionLeft,     // First expression (or null) given as SqlExpression instance
        tokenOperator,      // SqlBasisToken instance representing operation name
        expressionRight     // Second expression given as SqlExpression instance
    ) {
        super(expressionLeft ? expressionLeft.tokenStart : tokenOperator, expressionRight.tokenEnd);
        if (expressionLeft)
            this.expressionLeft = expressionLeft;
        this.expressionRight = expressionRight;
        this.tokenOperator = tokenOperator;
    }

    /**
    * Calculate value taking scope in account
    * @return mixed     Value
    */
    calcValue(
        scope       // Available tables and so on
    ) {
        function ensureNumber(value) {
            if (value === null || value === undefined || value === Infinity || value === -Infinity || value != value)
                return null;
            if (typeof value != 'number')
                return (value == '') ? 0 : 1;
            return value;
        }
        const left = this.expressionLeft ? this.expressionLeft.calcValue(scope) : undefined;
        const right = this.expressionRight.calcValue(scope);
        if (right === null)
            return null;
        const operator = this.tokenOperator.content.toUpperCase();
        switch (operator) {
            case '~':
                return ~right;
            case 'NOT':
                return right ? 0 : 1;
        }
        if (left === null)
            return null;
        switch (operator) {
            case '||':
                return '' + left + right;
            case '*':
                return ensureNumber(left * right);
            case '/':
                return ensureNumber(left / right);
            case '%':
                return ensureNumber(left % right);
            case '+':
                if (left === undefined)
                    return ensureNumber(value);
                return left + right;
            case '-':
                if (left === undefined)
                    return -ensureNumber(value);
                return ensureNumber(ensureNumber(left) - ensureNumber(right));
            case 'AND':
                return ensureNumber(left && right);
            case 'OR':
                return ensureNumber(left || right);
            case '<>':
            case '!=':
                return ensureNumber(left != right);
            case '<':
                return ensureNumber(left < right);
            case '>':
                return ensureNumber(left > right);
            case '<=':
                return ensureNumber(left <= right);
            case '>=':
                return ensureNumber(left >= right);
            case '=':
                return ensureNumber(left == right);
            case 'LIKE':
            case 'NOT LIKE':
                const regularExpression = SqlExpressionOperation.patternToRegExp(
                    this.expressionMiddle ? this.expressionMiddle.calcValue(scope) : right,
                    this.expressionMiddle ? right : '='
                );
                if (!regularExpression)
                    throw new SqlError('Ausdruck nach ESCAPE muss ein einzelnes Zeichen sein');
                return (!('' + left).match(regularExpression) != (operator == 'LIKE')) ? 1 : 0;
            case 'BETWEEN':
                return ensureNumber(left >= this.expressionMiddle.calcValue(scope) && left <= right);
            case 'NOT BETWEEN':
                return ensureNumber(left < this.expressionMiddle.calcValue(scope) || left > right);
            /*!*/
        }
throw new SqlError('Operation nicht implementiert', this);
        return null;
    }

    /**
    * Get name of SQL expression (i. e. attribute name if used as an attribute or table name)
    * @return String        Attribute name
    */
    getName() {
        //return (this.expressionLeft ? this.expressionLeft.getName() : '') + this.tokenOperator.content + this.expressionRight.getName();
        return this.toString();
    }

    /**
    * Get string representation of expression
    * @return String    String beeing valid SQL
    */
    toString() {
        return '' + (this.expressionLeft ? (this.expressionLeft.toString() + ' ') : '') + this.tokenOperator.content + ' ' + this.expressionRight.toString() + '';
    }

    /**
    * Get execution order of this operation
    * @return number        Priority (lower number means this operation is executed first)
    */
    getPriority() {
        return {
            '~' : 1,
            '*': 2, '/': 2, '%': 2,
            '+': 3, '-': 3, '&': 3, '^': 3, '||': 3,
            '=': 4, '>': 4, '<': 4, '>=': 4, '<=': 4, '<>': 4, '!=': 4, 'LIKE': 4, 'NOT LIKE' : 4, 'BETWEEN': 4, 'NOT BETWEEN' : 4,
            'NOT': 5,
            'AND': 6,
            'OR': 7
        }[this.tokenOperator.content.toUpperCase()] || Infinity;
    }

    /**
    * Eventually pass down this node the expression tree to fit operation priorities
    * @return SqlExpressionOperation    New root of sub tree former started with `this` beeing root
    */
    correctByPriority() {
        if (!(this.expressionLeft instanceof SqlExpressionOperation))
            return this;
        if (this.expressionLeft.getPriority() <= this.getPriority())
            return this;
        const root = this.expressionLeft;
        this.expressionLeft = this.expressionLeft.expressionRight;
        this.tokenStart = this.expressionLeft.tokenStart;
        root.expressionRight = this;
        root.tokenEnd = this.tokenEnd;
        this.correctByPriority();
        return root;
    }

    /**
    * Add optional part to bottom most possible expression of this part tree
    * @return boolean       Optional part added?
    */
    integrateOptionalPart(
        keyword,        // Keyword preceding optional part (upper case!)
        expression      // SqlExpression representing optional part
    ) {
        if (
            this.expressionRight instanceof SqlExpressionOperation
            && this.expressionRight.integrateOptionalPart(keyword, expression)
        )
            return true;
        if (this.expressionMiddle)
            return false;
        switch (keyword) {
            case 'ESCAPE':
                if (['LIKE', 'NOT LIKE'].indexOf(this.tokenOperator.content.toUpperCase()) == -1)
                    break;
                this.expressionMiddle = this.expressionRight;
                this.expressionRight = expression;
                this.tokenEnd = expression.tokenEnd;
                return true;
        }
        return false;
    }
}

// SqlExpression instance describing left side of operator (if there is any)
SqlExpressionOperation.prototype.expressionLeft = null;

// SqlExpression instance describing right side of operator
SqlExpressionOperation.prototype.expressionRight = null;

// Eventually used further SqlExpression instance (optional for LIKE operator and neccessary for BETWEEN operator)
SqlExpressionOperation.prototype.expressionMiddle = null;

/**
* Convert SQL LIKE pattern to RegExp
* @return RegExp    Regular expression given by SQL notation
*/
SqlExpressionOperation.patternToRegExp = function(
    pattern,    // SQL string with '_' meaning single char and '%' meaning any number of chars
    escape      // Optional: Char used in `pattern` to escape '_', '%' and itself (defaults to '=')
) {
    if (!(typeof escape == 'string') && !(escape instanceof String))
        escape = '=';
    escape = '' + escape;
    if (escape.length != 1)
        return null;
    pattern = '' + pattern;
    const special = '\\[]^.|$?()*+{}';
    let s = '';
    for (let n = 0; n < pattern.length; ++n) {
        let nextChar = pattern.charAt(n);
        switch(nextChar) {
            case '_':
                s += '.';
                continue;
            case '%':
                s += '.*';
                continue;
            case escape:
                ++n;
                nextChar = pattern.charAt(n);
                break;
        }
        s += ((special.indexOf(nextChar) > -1) ? '\\' : '') + nextChar;
    }
    return new RegExp('^' + s + '$');
}

/**
* Extend `expression.expressionRight` until there is keyword `keyword`, use following expression as `expression.expressionRight
* @return boolean       Found keyword?
*/
SqlExpressionOperation.addOperandThree = function(
    expression,     // SqlExpressionOperation instance to eventually add `expressionMiddle` attribute value
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    keyword         // Required content of token of type TOKEN_KEYWORD preceding third part of operation given in uppercase letters
) {
    while (
        arrTokens.length
        && (arrTokens[0].type != SqlBasisToken.TOKEN_KEYWORD || arrTokens[0].content.toUpperCase() != keyword)
    ) {
        const lengthBefore = arrTokens.length;
        expression.expressionRight = SqlExpressionOperation.extend(expression.expressionRight, arrTokens, true);
        if (arrTokens.length >= lengthBefore)
            break;
    }
    if (
        !arrTokens.length
        || arrTokens[0].type != SqlBasisToken.TOKEN_KEYWORD
        || arrTokens[0].content.toUpperCase() != keyword
    )
        throw new SqlError(keyword + ' erwartet', arrTokens[0]);
    const tokenKeyword = arrTokens.shift();
    const operandThree = [];
    if (!SqlExpression.fromTokens(arrTokens, operandThree, true) || !operandThree.length)
        throw new SqlError('Nach ' + keyword + ' muss ein Ausdruck folgen', tokenKeyword);
    expression.expressionMiddle = expression.expressionRight;
    expression.expressionRight = operandThree[0];
    expression.tokenEnd = expression.expressionRight.tokenEnd;
}

/**
* Extend a part of an expression by one more operation if possible
* @return SqlExpressionOperation        New root element of expression tree
*/
SqlExpressionOperation.extend = function(
    expressionLeft, // SqlExpression instance to extend
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    noRecursion     // Optional: Set `true` to stop recursion calls of `SqlExpressionOperation.extend`
) {
    if (!arrTokens.length)
        return expressionLeft;
    if (!(arrTokens[0].type == SqlBasisToken.TOKEN_KEYWORD && ['NOT', 'AND', 'OR', 'LIKE', 'ESCAPE', 'BETWEEN'].indexOf(arrTokens[0].content.toUpperCase()) != -1)
        && !(arrTokens[0].type == SqlBasisToken.TOKEN_OPERATOR && SqlBasisToken.OPERATORS.indexOf(arrTokens[0].content) != -1))
        return expressionLeft;
    let tokenOperator = arrTokens.shift();
    if (
        tokenOperator.content.toUpperCase() == 'NOT'
        && arrTokens.length
        && ['LIKE', 'BETWEEN'].indexOf(arrTokens[0].content.toUpperCase()) > -1
    ) {
        const tokenNot = tokenOperator;
        const tokenType = arrTokens.shift();
        tokenOperator = new SqlBasisToken(SqlBasisToken.TOKEN_KEYWORD, tokenNot.content + ' ' + tokenType.content, tokenNot.line, tokenNot.offset);
    }
    const operator = tokenOperator.content.toUpperCase();
    const operand = [];
    if (!SqlExpression.fromTokens(arrTokens, operand, true) || !operand.length)
        throw new SqlError('Nach dem Operator ' + operator + ' muss ein Ausdruck folgen', tokenOperator);
    if (['ESCAPE'].indexOf(operator) != -1) {
        if (!(expressionLeft instanceof SqlExpressionOperation) || !expressionLeft.integrateOptionalPart(operator, operand[0]))
            throw new SqlError('Unklar, worauf sich ' + operator + ' bezieht', tokenOperator);
        return noRecursion ? expressionLeft : SqlExpressionOperation.extend(expressionLeft, arrTokens);
    }
    const expressionCreated = new SqlExpressionOperation(expressionLeft, tokenOperator, operand[0]);
    if (['BETWEEN', 'NOT BETWEEN'].indexOf(tokenOperator.content.toUpperCase()) != -1)
        SqlExpressionOperation.addOperandThree(expressionCreated, arrTokens, 'AND');
    const expressionParent = expressionCreated.correctByPriority();
    return noRecursion ? expressionParent : SqlExpressionOperation.extend(expressionParent, arrTokens);
}

/**
* Read from token list and convert first token(s) to SqlExpression instance if possible
* @return boolean       Token(s) converted?
*/
SqlExpressionOperation.fromTokens = function(
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    arrParts,       // Array to output created SqlExpression instance
    uncompleteOps   // Optional: Set true if a maximum of a single operation is to be read
) {
    if (!arrTokens.length || (!(arrTokens[0].type == SqlBasisToken.TOKEN_KEYWORD && arrTokens[0].content.toUpperCase() == 'NOT') && !(arrTokens[0].type == SqlBasisToken.TOKEN_OPERATOR && arrTokens[0].content == '~')))
        return false;
    const tokenOperator = arrTokens.shift();
    const operand = [];
    if (!SqlExpression.fromTokens(arrTokens, operand, true) || !operand.length)
        throw new SqlError('Nach dem Operator wird ein Ausdruck erwartet', tokenOperator);
    let expression = new SqlExpressionOperation(null, tokenOperator, operand[0]);
    if (!uncompleteOps)
        expression = SqlExpressionOperation.extend(expression, arrParts);
    arrParts.push(expression);
    return true;
}

/***[ SqlExpressionFromClause.js ] *************************/
/** Class for a SQL FROM clause within a SQL statement */
class SqlExpressionFromClause extends SqlExpression {
    constructor(
        tokenStart,     // SqlBasisToken of type SqlBasisToken.TOKEN_KEYWORD having content 'SELECT'
        tables          // Array of SqlExpressionTableImport instances
    ) {
        super(tokenStart, tables.length ? tables[tables.length - 1].tokenEnd : undefined);
        this.tables = tables;
    }

    /**
    * Calculate value taking scope in account
    * @return mixed     Value
    */
    calcValue(
        scope       // Available tables and so on
    ) {
        scope.startNextTable();
        const names = this.tables.map(table => table.getName());
        const duplicate = names.reduce((prev, item, index, arr) => (prev || ((arr.indexOf(item) != index) ? item : false)), false);
        if (duplicate)
            throw new SqlError('Tabellenname ' + duplicate + ' ist nicht eindeutig', this);
        this.tables.forEach((table, index) => {
            const forkedScope = scope.clone();
            table.calcValue(forkedScope);
            scope.tableResult = index ? scope.tableResult.crossMultiply(forkedScope.tableResult) : forkedScope.tableResult;
        });
        return undefined;
    }

    /**
    * Get name of SQL expression (i. e. attribute name if used as an attribute or table name)
    * @return String        Attribute name
    */
    getName() {
        return 'FROM ' + this.tables.map(att => att.getName()).join(', ');
    }

    /**
    * Get string representation of expression
    * @return String    String beeing valid SQL
    */
    toString() {
        return 'FROM ' + this.tables.map(att => att.toString()).join(', ');
    }
}

// Array of SqlExpressionSelectAttribute instances
SqlExpressionFromClause.prototype.tables = [];

/**
* Read from token list and convert first token(s) to SqlExpression instance if possible
* @return boolean       Token(s) converted?
*/
SqlExpressionFromClause.fromTokens = function(
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    arrParts        // Array to output created SqlExpression instance
) {
    if (!arrTokens.length || arrTokens[0].type != SqlBasisToken.TOKEN_KEYWORD || arrTokens[0].content.toUpperCase() != 'FROM')
        return false;
    const tables = [];
    const tokenStart = arrTokens[0];
    let token;
    while (arrTokens.length) {
        if (tables.length && (arrTokens[0].type != SqlBasisToken.TOKEN_SEPARATOR || arrTokens[0].content != ','))
            break;
        token = arrTokens.shift();
        if (!SqlExpressionTableImport.fromTokens(arrTokens, tables))
            throw new SqlError('Nach ' + token.toString() + ' muss ein Tabellenname folgen', token);
    }
    arrParts.push(new SqlExpressionFromClause(tokenStart, tables));
    return true;
}

/***[ SqlExpressionWhereClause.js ] *************************/
/** Class for a SQL WHERE clause within a SQL statement */
class SqlExpressionWhereClause extends SqlExpression {
    constructor(
        tokenStart,     // SqlBasisToken of type SqlBasisToken.TOKEN_KEYWORD having content 'SELECT'
        expression      // SqlExpression instance describing content of WHERE clause
    ) {
        super(tokenStart, expression.tokenEnd);
        this.expression = expression;
    }

    /**
    * Calculate value taking scope in account
    * @return mixed     Value
    */
    calcValue(
        scope       // Available tables and so on
    ) {
        const tableSource = scope.tableResult;
        scope.startNextTable();
        const forkedScope = scope.clone();
        forkedScope.startNextTable();
        forkedScope.tableResult = tableSource.clone();
        scope.tableResult = tableSource.clone();
        scope.tableResult.data = tableSource.data.filter(row => {
            forkedScope.tableResult.data = [row];
            return this.expression.calcValue(forkedScope);
        });
        return undefined;
    }

    /**
    * Get string representation of expression
    * @return String    String beeing valid SQL
    */
    toString() {
        return 'WHERE ' + this.expression.toString();
    }
}

// Array of SqlExpressionSelectAttribute instances
SqlExpressionWhereClause.prototype.tables = [];

/**
* Read from token list and convert first token(s) to SqlExpression instance if possible
* @return boolean       Token(s) converted?
*/
SqlExpressionWhereClause.fromTokens = function(
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    arrParts        // Array to output created SqlExpression instance
) {
    if (!arrTokens.length || arrTokens[0].type != SqlBasisToken.TOKEN_KEYWORD || arrTokens[0].content.toUpperCase() != 'WHERE')
        return false;
    const tokenStart = arrTokens.shift();
    const index = arrParts.length;
    if (!SqlExpression.fromTokens(arrTokens, arrParts) || !arrParts[index])
        throw new SqlError('Nach WHERE muss ein gültiger Ausdruck folgen', tokenStart);
    arrParts[index] = new SqlExpressionWhereClause(tokenStart, arrParts[index]);
    return true;
}

/***[ SqlExpressionSelectAttribute.js ] *************************/
/** Class for an attribute within a SQL SELECT statement */
class SqlExpressionSelectAttribute extends SqlExpression {
    constructor(
        expression,     // SqlExpression instance describing derivation of attribute's value
        tokenName       // Optional: SqlBasisToken instance having type SqlBasisToken.TOKEN_QUOTEDNAME or SqlBasisToken.TOKEN_NORMAL describing attribute's name
    ) {
        super(expression.tokenStart, tokenName ? tokenName : expression.tokenEnd);
        this.expression = expression;
        if (tokenName)
            this.tokenName = tokenName;
    }

    /**
    * Calculate value taking scope in account
    * @return mixed     Value
    */
    calcValue(
        scope       // Available tables and so on
    ) {
        return this.expression.calcValue(scope);
    }

    /**
    * Get name of SQL expression (i. e. attribute name if used as an attribute or table name)
    * @return String        Attribute name
    */
    getName() {
        return this.tokenName ? this.tokenName.content : this.expression.getName();
    }

    /**
    * Get string representation of expression
    * @return String    String beeing valid SQL
    */
    toString() {
        return this.expression.toString() + (this.tokenName ? (' AS ' + this.tokenName.toString()) : '');
    }
}

// SqlExpression instance describing derivation of attribute's value
SqlExpressionSelectAttribute.prototype.expression = null;

// SqlBasisToken instance having type SqlBasisToken.TOKEN_QUOTEDNAME or SqlBasisToken.TOKEN_NORMAL describing attribute's name (if any)
SqlExpressionSelectAttribute.prototype.tokenName = null;

/**
* Read from token list and convert first token(s) to SqlExpression instance if possible
* @return boolean       Token(s) converted?
*/
SqlExpressionSelectAttribute.fromTokens = function(
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    arrParts        // Array to output created SqlExpression instance
) {
    const arrExpression = [];
    if (!SqlExpression.fromTokens(arrTokens, arrExpression))
        return false;
    const expression = arrExpression[0];
    if (arrTokens.length && arrTokens[0].type == SqlBasisToken.TOKEN_KEYWORD && arrTokens[0].content.toUpperCase() == 'AS') {
        const tokenAs = arrTokens.shift();
        if (!arrTokens.length || [SqlBasisToken.TOKEN_QUOTEDNAME, SqlBasisToken.TOKEN_NORMAL].indexOf(arrTokens[0].type) == -1)
            throw new SqlError('Attributname nach AS erwartet', tokenAs);
        arrParts.push(new SqlExpressionSelectAttribute(expression, arrTokens.shift()));
        return true;
    }
    if (arrTokens.length && [SqlBasisToken.TOKEN_QUOTEDNAME, SqlBasisToken.TOKEN_NORMAL].indexOf(arrTokens[0].type) != -1) {
        arrParts.push(new SqlExpressionSelectAttribute(expression, arrTokens.shift()));
        return true;
    }
    arrParts.push(new SqlExpressionSelectAttribute(expression));
    return true;
}

/***[ SqlExpressionSelectAll.js ] *************************/
/** Class for an "select all or select all attributes of table" within a SQL SELECT clause */
class SqlExpressionSelectAll extends SqlExpression {
    constructor(
        tokenAsterix,   // SqlBasisToken instance having content '*'
        tokenTable      // Optional: SqlBasisToken instance having type SqlBasisToken.TOKEN_QUOTEDNAME or SqlBasisToken.TOKEN_NORMAL describing table's name
    ) {
        super(tokenTable || tokenAsterix, tokenAsterix);
        if (tokenTable)
            this.tokenTable = tokenTable
    }

    /**
    * Create SqlExpressionSelectAttribute like objects for each fitting table row and add to `arrOut`
    * @return undefined
    */
    expand(
        scope,      // Available tables and so on
        arrOut      // Array to add pseudo expressions to
    ) {
        const cols = scope.tableResult.findCol(undefined, this.tokenTable ? this.tokenTable.content : undefined);
        cols.forEach(({ name, index }) => arrOut.push({
            getName: () => name,
            calcValue: (scope) => scope.tableResult.data[0][index]
        }));
    }

    /**
    * Get name of SQL expression (i. e. attribute name if used as an attribute or table name)
    * @return String        Attribute name
    */
    getName() {
        return (this.tokenTable ? (this.tokenTable.toString() + '.') : '') + '*';
    }

    /**
    * Get string representation of expression
    * @return String    String beeing valid SQL
    */
    toString() {
        return this.getName();
    }
}

// SqlBasisToken instance having type SqlBasisToken.TOKEN_QUOTEDNAME or SqlBasisToken.TOKEN_NORMAL describing table's name
SqlExpressionSelectAll.prototype.tokenTable = null;

/**
* Read from token list and convert first token(s) to SqlExpression instance if possible
* @return boolean       Token(s) converted?
*/
SqlExpressionSelectAll.fromTokens = function(
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    arrParts        // Array to output created SqlExpression instance
) {
    if (arrTokens.length && arrTokens[0].type == SqlBasisToken.TOKEN_OPERATOR && arrTokens[0].content == '*') {
        arrParts.push(new SqlExpressionSelectAll(arrTokens.shift()));
        return true;
    }
    if (
        arrTokens.length >= 3
        && [SqlBasisToken.TOKEN_QUOTEDNAME, SqlBasisToken.TOKEN_NORMAL].indexOf(arrTokens[0].type) != -1
        && arrTokens[1].type == SqlBasisToken.TOKEN_SEPARATOR && arrTokens[1].content == '.'
        && arrTokens[2].type == SqlBasisToken.TOKEN_OPERATOR && arrTokens[2].content == '*'
    ) {
        arrParts.push(new SqlExpressionSelectAll(arrTokens[2], arrTokens[0]));
        arrTokens.splice(0, 3);
        return true;
    }
    return false;
}

/***[ SqlExpressionSelectClause.js ] *************************/
/** Class for a SQL SELECT statement (only SELECT clause!) */
class SqlExpressionSelectClause extends SqlExpression {
    constructor(
        tokenStart,     // SqlBasisToken of type SqlBasisToken.TOKEN_KEYWORD having content 'SELECT'
        attributes      // Array of SqlExpressionSelectAttribute instances
    ) {
        super(tokenStart, attributes.length ? attributes[attributes.length - 1].tokenEnd : undefined);
        this.attributes = attributes;
    }

    /**
    * Calculate value taking scope in account
    * @return mixed     Value
    */
    calcValue(
        scope       // Available tables and so on
    ) {
        const attributes = [];
        this.attributes.forEach(att => {
            if (att instanceof SqlExpressionSelectAttribute)
                return attributes.push(att);
            att.expand(scope, attributes);
        });
        const tableInput = scope.tableResult;
        scope.startNextTable();
        const forkedScope = scope.clone();
        scope.tableResult.atts = attributes.map(att => ({ name: att.getName(), tablename: undefined }));
        forkedScope.startNextTable();
        forkedScope.tableResult.atts = tableInput.atts;
        scope.tableResult.data = tableInput.data.map(row => {
            forkedScope.tableResult.data = [row];
            return attributes.map(att => att.calcValue(forkedScope));
        });
        if (this.isDistinct)
            scope.tableResult.data = scope.tableResult.data.filter((row, index, arr) => arr.slice(0, index).reduce((prev, arr) => prev && !arr.reduce((gleich, item, i) => gleich && item == row[i], true), true));
        return undefined;
    }

    /**
    * Get name of SQL expression (i. e. attribute name if used as an attribute or table name)
    * @return String        Attribute name
    */
    getName() {
        return 'SELECT ' + this.attributes.map(att => att.getName()).join(', ');
    }

    /**
    * Get string representation of expression
    * @return String    String beeing valid SQL
    */
    toString() {
        return 'SELECT ' + (this.isDistinct ? 'DISTINCT ' : ((this.isDistinct === undefined) ? '' : 'ALL ')) + this.attributes.map(att => att.toString()).join(', ');
    }
}

// Is DISTINCT (true), ALL (false) or nothing (undefined) specified?
SqlExpressionSelectClause.prototype.isDistinct = undefined;

// Array of SqlExpressionSelectAttribute instances
SqlExpressionSelectClause.prototype.attributes = [];

/**
* Read from token list and convert first token(s) to SqlExpression instance if possible
* @return boolean       Token(s) converted?
*/
SqlExpressionSelectClause.fromTokens = function(
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    arrParts        // Array to output created SqlExpression instance
) {
    if (!arrTokens.length || arrTokens[0].type != SqlBasisToken.TOKEN_KEYWORD || arrTokens[0].content.toUpperCase() != 'SELECT')
        return false;
    const attributes = [];
    const tokenStart = arrTokens[0];
    let type = undefined;
    if (arrTokens.length > 1 && arrTokens[1].type == SqlBasisToken.TOKEN_KEYWORD && ['ALL', 'DISTINCT'].indexOf(arrTokens[1].content.toUpperCase()) != -1)
        type = (arrTokens.splice(1, 1)[0].content.toUpperCase() == 'DISTINCT');
    let token;
    while (arrTokens.length) {
        if (attributes.length && (arrTokens[0].type != SqlBasisToken.TOKEN_SEPARATOR || arrTokens[0].content != ','))
            break;
        token = arrTokens.shift();
        if (!SqlExpressionSelectAttribute.fromTokens(arrTokens, attributes) && !SqlExpressionSelectAll.fromTokens(arrTokens, attributes))
            throw new SqlError('Nach ' + token.toString() + ' muss ein Attribut folgen', token);
    }
    arrParts.push(new SqlExpressionSelectClause(tokenStart, attributes));
    if (type !== undefined)
        arrParts[arrParts.length - 1].isDistinct = type;
    return true;
}

/***[ SqlExpressionOrderByRule.js ] *************************/
/** Class for an sorting rule within a SQL ORDER BY statement */
class SqlExpressionOrderByRule extends SqlExpression {
    constructor(
        expression,     // SqlExpression instance describing derivation of attribute's value
        isDescending,   // Optional: Descending order? (undefined → implicit ASC, false → explicit ASC, true → DESC)
        tokenEnd        // Optional: Last token
    ) {
        super(expression.tokenStart, tokenEnd ? tokenEnd : expression.tokenEnd);
        this.expression = expression;
        if (isDescending !== undefined)
            this.isDescending = isDescending ? true : false;
    }

    /**
    * Calculate value taking scope in account
    * @return mixed     Value
    */
    calcValue(
        scopeA,     // Available tables and so on
        scopeB,     // Available tables and so on
    ) {
        let valueA = this.expression.calcValue(this.isDescending ? scopeB : scopeA);
        let valueB = this.expression.calcValue(this.isDescending ? scopeA : scopeB);
        if (valueA instanceof Date)
            valueA = valueA.getTime();
        if (valueB instanceof Date)
            valueB = valueB.getTime();
        if (valueA === null || valueA === undefined || valueA != valueA)
            return -1;
        if (valueB === null || valueB === undefined || valueB != valueB)
            return 1;
        if (Number.isFinite(valueA) && Number.isFinite(valueB))
            return valueA - valueB;
        return ('' + valueA).localeCompare(valueB);
    }

    /**
    * Get name of SQL expression (i. e. attribute name if used as an attribute or table name)
    * @return String        Attribute name
    */
    getName() {
        return this.expression.getName();
    }

    /**
    * Get string representation of expression
    * @return String    String beeing valid SQL
    */
    toString() {
        return this.expression.toString() + ((this.isDescending === undefined) ? '' : (this.isDescending ? ' DESC' : ' ASC'));
    }
}

// SqlExpression instance describing derivation of attribute's value
SqlExpressionOrderByRule.prototype.expression = null;

// Is this a descending order? (undefined → neither ASC nor DESC given, true → explicit ASC, false → DESC)
SqlExpressionOrderByRule.prototype.isDescending = undefined;

/**
* Read from token list and convert first token(s) to SqlExpression instance if possible
* @return boolean       Token(s) converted?
*/
SqlExpressionOrderByRule.fromTokens = function(
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    arrParts        // Array to output created SqlExpression instance
) {
    const arrExpression = [];
    if (!SqlExpression.fromTokens(arrTokens, arrExpression))
        return false;
    const expression = arrExpression[0];
    if (arrTokens.length && arrTokens[0].type == SqlBasisToken.TOKEN_KEYWORD && ['ASC', 'DESC'].indexOf(arrTokens[0].content.toUpperCase()) != -1) {
        arrParts.push(new SqlExpressionOrderByRule(expression, arrTokens[0].content.toUpperCase() == 'DESC', arrTokens.shift()));
        return true;
    }
    arrParts.push(new SqlExpressionOrderByRule(expression));
    return true;
}

/***[ SqlExpressionOrderByClause.js ] *************************/
/** Class for a SQL ORDER BY clause within a SQL statement */
class SqlExpressionOrderByClause extends SqlExpression {
    constructor(
        tokenStart,     // SqlBasisToken of type SqlBasisToken.TOKEN_KEYWORD having content 'SELECT'
        arrRules        // SqlExpressionOrderByRule instances describing sorting order
    ) {
        super(tokenStart, arrRules[arrRules.length - 1].tokenEnd);
        this.arrRules = arrRules.slice();
    }

    /**
    * Calculate value taking scope in account
    * @return mixed     Value
    */
    calcValue(
        scope       // Available tables and so on
    ) {
        const tableSource = scope.tableResult;
        scope.startNextTable();
        const forkA = scope.clone();
        forkA.startNextTable();
        forkA.tableResult = tableSource.clone();
        const forkB = scope.clone();
        forkB.startNextTable();
        forkB.tableResult = tableSource.clone();
        scope.tableResult = tableSource.clone();
        scope.tableResult.data.sort((rowA, rowB) => this.arrRules.reduce((prev, rule) => {
            if (prev)
                return prev;
            forkA.tableResult.data = [rowA];
            forkB.tableResult.data = [rowB];
            return rule.calcValue(forkA, forkB);
        }, 0));
        return undefined;
    }

    /**
    * Get string representation of expression
    * @return String    String beeing valid SQL
    */
    toString() {
        return 'ORDER BY ' + this.arrRules.map(rule => rule.toString()).join(', ');
    }
}

// Array of SqlExpressionOrderByRule instances describing order rules
SqlExpressionOrderByClause.prototype.arrRules = [];

/**
* Read from token list and convert first token(s) to SqlExpression instance if possible
* @return boolean       Token(s) converted?
*/
SqlExpressionOrderByClause.fromTokens = function(
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    arrParts        // Array to output created SqlExpression instance
) {
    if (!arrTokens.length || arrTokens[0].type != SqlBasisToken.TOKEN_KEYWORD || arrTokens[0].content.toUpperCase() != 'ORDER')
        return false;
    const tokenStart = arrTokens.shift();
    if (!arrTokens.length || arrTokens[0].type != SqlBasisToken.TOKEN_KEYWORD || arrTokens[0].content.toUpperCase() != 'BY')
        throw new SqlError('Nach ORDER muss das Schlüsselwort BY folgen', tokenStart);
    const tokenBy = arrTokens.shift();
    let tokenLast = tokenBy;
    const arrRules = [];
    while (true) {
        const xtra = [];
        if (!SqlExpressionOrderByRule.fromTokens(arrTokens, xtra) || !xtra.length)
            throw new SqlError((arrRules.length ? 'Nach einem Komma muss eine weitere' : 'Nach ORDER BY muss eine') + ' gültige Sortierbedingung folgen', tokenLast);
        arrRules.push(xtra[0]);
        if (!arrTokens.length || arrTokens[0].type != SqlBasisToken.TOKEN_SEPARATOR || arrTokens[0].content != ',')
            break;
        tokenLast = arrTokens.shift();
    }
    arrParts.push(new SqlExpressionOrderByClause(tokenStart, arrRules));
    return true;
}

/***[ SqlExpressionSelect.js ] *************************/
/** Class for a whole SQL SELECT command */
class SqlExpressionSelect extends SqlExpression {
    constructor(
        select,     // SqlExpressionSelectClause instance representing SELECT clause
        from,       // Optional: SqlExpressionFromClause instance representing FROM clause
        where,      // Optional: SqlExpressionWhereClause instance representing WHERE clause
        order       // Optional: SqlExpressionOrderByClause instance representing ORDER clause
    ) {
        const lastClause = order || where || from || select;
        super(select.tokenStart, lastClause.tokenEnd);
        this.select = select;
        if (from)
            this.from = from;
        if (where)
            this.where = where;
        if (order)
            this.order = order;
    }

    /**
    * Calculate value taking scope in account
    * @return mixed     Value
    */
    calcValue(
        scope       // Available tables and so on
    ) {
        if (this.from)
            this.from.calcValue(scope);
        else {
            scope.startNextTable();
            scope.tableResult.data.push([]);
        }
        if (this.where)
            this.where.calcValue(scope);
        if (this.order)
            this.order.calcValue(scope);
        this.select.calcValue(scope);
        return undefined;
    }

    /**
    * Get name of SQL expression (i. e. attribute name if used as an attribute or table name)
    * @return String        Attribute name
    */
    getName() {
        return 'SELECT ' + this.attributes.map(att => att.getName()).join(', ');
    }

    /**
    * Get string representation of expression
    * @return String    String beeing valid SQL
    */
    toString() {
        return [this.select, this.from, this.where, this.order].filter(obj => obj).map(expr => expr.toString()).join(' ')
    }
}

// SqlExpressionSelectClause instance representing SELECT clause
SqlExpressionSelect.prototype.select = null;

// SqlExpressionFromClause instance representing FROM clause
SqlExpressionSelect.prototype.from = null;

// SqlExpressionWhereClause instance representing WHERE clause
SqlExpressionSelect.prototype.where = null;

// SqlExpressionOrderClause instance representing ORDER clause
SqlExpressionSelect.prototype.order = null;

/**
* Read from token list and convert first token(s) to SqlExpression instance if possible
* @return boolean       Token(s) converted?
*/
SqlExpressionSelect.fromTokens = function(
    arrTokens,      // List oft SqlBasisToken instances (without comments and whitespace) to be converted
    arrParts        // Array to output created SqlExpression instance
) {
    const parts = [];
    if (!SqlExpressionSelectClause.fromTokens(arrTokens, parts))
        return false;
    SqlExpressionFromClause.fromTokens(arrTokens, parts);
    SqlExpressionWhereClause.fromTokens(arrTokens, parts);
    SqlExpressionOrderByClause.fromTokens(arrTokens, parts);
    arrParts.push(new SqlExpressionSelect(
        parts.shift(),
        (parts.length && (parts[0] instanceof SqlExpressionFromClause)) ? parts.shift() : null,
        (parts.length && (parts[0] instanceof SqlExpressionWhereClause)) ? parts.shift() : null,
        (parts.length && (parts[0] instanceof SqlExpressionOrderByClause)) ? parts.shift() : null
    ));
    return true;
}

/***[ SqlQueryTerminal.js ] *************************/
/** Class for user access to sql query input and execution */
function SqlQueryTerminal() {
    this.node = createHtml('div', null, { 'class': 'SqlQueryTerminal' });
    this._msgPlace = new ScoTools.MessagePlace();
    this._msgPlace.appendAsChild(this.node);
    this._nodeOutput = createHtml('div', this.node, { 'class': 'SqlQueryTerminalOutput' });
    this._editor = new CodeMirror(this.node, {lineNumbers: false, mode: "text/x-sql", theme: "default", viewportMargin: Infinity, indentUnit: 4, lineWrapping: true });
    this._editor.on('change', () => this._procCodeChange());
    const button = createHtml('button', this.node, { type: 'button' }, '⚙️ Abfrage ausführen');
    button.addEventListener('click', () => this.execute());
}

// SqlScope instance containg all available tables
SqlQueryTerminal.prototype.scope = null;

// Is there a results table beeing displayed?
SqlQueryTerminal.prototype._displayResultsTable = false;

// HTML to display if there ist no results table to show
SqlQueryTerminal.prototype._hint = '';
Object.defineProperty(SqlQueryTerminal.prototype, 'hint', {
    enumerable: true,
    get() { return this._hint; },
    set(html) {
        this._hint = html;
        if (!this._displayResultsTable)
            this._nodeOutput.innerHTML = html;
    }
});

// Inputed SQL code
Object.defineProperty(SqlQueryTerminal.prototype, 'code', {
    enumerable: true,
    get() { return this._editor.getValue(); },
    set(str) { this._editor.setValue(str); }
});

/** Function to ask, whether or not to run a code; has to return either true for yes and false for no or an object like { accept: "Execution allowed?", token: "SqlBasisToken or SqlExpression instance representing problem (optional)", msg: "Error message to show (optional)" } */
SqlQueryTerminal.prototype.onbeforeExecution = function(
    code,       // Value of this.code
    tokens,     // Array of all tokens representing code
    compiled,   // SqlExpression instance to execute
) {
    return true;
}

/** Function to call whenever program execution terminates; if function returns either false for a general missmatch or an object like { accept: "Was execution a success?", token: "SqlBasisToken or SqlExpression instance representing problem (optional)", msg: "Message to display" } */
SqlQueryTerminal.prototype.onafterExecution = function(
    result,     // SqlTable instance representing results table
    code,       // Value of this.code
    tokens,     // Array of all tokens representing code
    compiled,   // SqlExpression instance to execute
) {
    return true;
}

/** Function to call whenever code input value changes */
SqlQueryTerminal.prototype.oncodeChange = function(
    code        // Value this.code
) {
}

/** Can code be edited? */
Object.defineProperty(SqlQueryTerminal.prototype, 'editingEnabled', {
    enumerable: true,
    get() { return !this._editor.isReadOnly(); },
    set(v) {
        this._editor.setOption('readOnly', !v)
    }
});

/**
* Remove all former messages from message output
* @return undefined
*/
SqlQueryTerminal.prototype.clearMsg = function() {
    this._msgPlace.clear();
}

/**
* Display message
* @return undefined
*/
SqlQueryTerminal.prototype.addMsg = function(
    text,       // Text of message
    type        // Type of message (optional, default value Message.TYPE_FAILED)
) {
    this._msgPlace.add(text, type);
}

/**
* React on code change
* @return undefined
*/
SqlQueryTerminal.prototype._procCodeChange = function() {
    if (this._displayResultsTable) {
        this._nodeOutput.innerHTML = this.hint;
        this._displayResultsTable = false;
    }
    if (typeof this.oncodeChange == 'function')
        this.oncodeChange.call(this, this.code);
}

/**
* Select a specific part of the inputed code
* @return undefined
*/
SqlQueryTerminal.prototype.selectPart = function(
    partStart,      // SqlBasisToken or SqlExpression instance representing start of selection
    partEnd         // Optional: SqlBasisToken or SqlExpression instance representing end of selection
) {
    if (!partEnd)
        partEnd = partStart;
    if (partStart instanceof SqlExpression)
        partStart = partStart.tokenStart;
    if (partEnd instanceof SqlExpression)
        partEnd = partEnd.tokenEnd || partEnd.tokenStart;
    const end = partEnd.getEnd();
    this._editor.setSelection({line: partStart.line - 1, ch: partStart.offset - 1}, {line: end[0] - 1, ch: end[1] + 1});
}

/**
* Check and execute inputed query
* @return undefined
*/
SqlQueryTerminal.prototype.execute = function() {
    const code = this.code;
    const tokens = SqlBasisToken
        .tokenize(code)
        .filter(token => [SqlBasisToken.TOKEN_WHITESPACE, SqlBasisToken.TOKEN_COMMENTLINE, SqlBasisToken.TOKEN_COMMENTMULTI].indexOf(token.type) == -1);
    const originalTokens = tokens.slice();
    const parts = [];
    try {
        if (!SqlExpressionSelect.fromTokens(tokens, parts) || !parts.length) {
            this.addMsg("Kein gültiger SELECT-Befehl");
            return;
        }
        while (tokens.length && tokens[0].content == ';' && tokens[0].type == SqlBasisToken.TOKEN_SEPARATOR)
            tokens.splice(0, 1);
        if (tokens.length) {
            this.selectPart(tokens[0], tokens[tokens.length - 1]);
            this.addMsg("Das Markierte konnte nicht erkannt werden");
            return;
        }
        const query = parts[0];
        if (typeof this.onbeforeExecution == 'function') {
            let allowExecute = this.onbeforeExecution.call(this, code, originalTokens, query);
            if (!(allowExecute instanceof Object))
                allowExecute = { accept: allowExecute };
            if (!allowExecute.accept) {
                this.addMsg(allowExecute.msg || 'Eingabe entspricht nicht den Anforderungen');
                if (allowExecute.token)
                    this.selectPart(allowExecute.token);
                return false;
            }
        }
        const scope = this.scope;
        query.calcValue(scope);
        this._nodeOutput.textContent = '';
        this._nodeOutput.appendChild(scope.tableResult.toDOM());
        this._displayResultsTable = true;
        // Perform level specific checks
        if (typeof this.onafterExecution == 'function') {
            let result = this.onafterExecution.call(this, scope.tableResult, code, originalTokens, query);
            if (!(result instanceof Object))
                result = { accept: result };
            if (!result.accept) {
                this.addMsg(result.msg || 'Ergebnistabelle entspricht nicht den Anforderungen');
                if (result.token)
                    this.selectPart(result.token);
                return false;
            }
            this.addMsg(result.msg || 'SQL-Abfrage ausgeführt', ScoTools.Message.TYPE_SUCCESS);
        }
    } catch(err) {
        if (!(err instanceof SqlError))
            throw err;
        this.addMsg("Fehler: " + err.message);
    }
}
